@@ -1,10 +1,11 @@
 /*
-  Copyright(C) 2012-2018  Brazil
-  Copyright(C) 2018-2020  Sutou Kouhei <kou@clear-code.com>
+  Copyright (C) 2012-2018  Brazil
+  Copyright (C) 2018-2023  Sutou Kouhei <kou@clear-code.com>
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
-  License version 2.1 as published by the Free Software Foundation.
+  License as published by the Free Software Foundation; either
+  version 2.1 of the License, or (at your option) any later version.
 
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -31,8 +32,10 @@
   Just for backward compatibility. See grn_plugin_charlen() instead.
  */
 int
-grn_tokenizer_charlen(grn_ctx *ctx, const char *str_ptr,
-                      unsigned int str_length, grn_encoding encoding)
+grn_tokenizer_charlen(grn_ctx *ctx,
+                      const char *str_ptr,
+                      unsigned int str_length,
+                      grn_encoding encoding)
 {
   return grn_plugin_charlen(ctx, str_ptr, str_length, encoding);
 }
@@ -41,8 +44,10 @@ grn_tokenizer_charlen(grn_ctx *ctx, const char *str_ptr,
   Just for backward compatibility. See grn_plugin_isspace() instead.
  */
 int
-grn_tokenizer_isspace(grn_ctx *ctx, const char *str_ptr,
-                      unsigned int str_length, grn_encoding encoding)
+grn_tokenizer_isspace(grn_ctx *ctx,
+                      const char *str_ptr,
+                      unsigned int str_length,
+                      grn_encoding encoding)
 {
   return grn_plugin_isspace(ctx, str_ptr, str_length, encoding);
 }
@@ -86,7 +91,8 @@ grn_tokenizer_have_tokenized_delimiter(grn_ctx *ctx,
 
   while ((char_length = grn_charlen_(ctx, current, end, encoding)) > 0) {
     if (grn_tokenizer_is_tokenized_delimiter(ctx,
-                                             current, char_length,
+                                             current,
+                                             (unsigned int)char_length,
                                              encoding)) {
       return GRN_TRUE;
     }
@@ -108,15 +114,20 @@ grn_tokenizer_query_ensure_normalized(grn_ctx *ctx, grn_tokenizer_query *query)
     grn_obj_close(ctx, query->normalized_query);
   }
   query->normalized_query = grn_string_open_(ctx,
-                                             query->ptr,
-                                             query->length,
+                                             query->data,
+                                             query->size,
                                              query->lexicon,
-                                             query->normalize_flags,
+                                             (int)(query->normalize_flags),
                                              query->encoding);
   if (!query->normalized_query) {
     query->have_tokenized_delimiter = GRN_FALSE;
-    GRN_PLUGIN_ERROR(ctx, GRN_TOKENIZER_ERROR,
-                     "[tokenizer][normalize] failed to open normalized string");
+    GRN_PLUGIN_ERROR(ctx,
+                     GRN_TOKENIZER_ERROR,
+                     "[tokenizer][normalize] failed to open normalized string: "
+                     "<%.*s>(%s)",
+                     (int)(query->size),
+                     query->data,
+                     grn_encoding_to_string(query->encoding));
     return;
   }
 
@@ -159,9 +170,9 @@ grn_tokenizer_query_init(grn_ctx *ctx, grn_tokenizer_query *query)
 {
   query->normalize_flags = 0;
   query->normalized_query = NULL;
-  query->query_buf = NULL;
+  query->data = NULL;
   query->ptr = NULL;
-  query->length = 0;
+  query->size = 0;
   query->flags = 0;
   query->tokenize_mode = GRN_TOKENIZE_ADD;
   query->token_mode = query->tokenize_mode;
@@ -177,11 +188,15 @@ grn_tokenizer_query_init(grn_ctx *ctx, grn_tokenizer_query *query)
 
   query->options = NULL;
 
+  query->domain = GRN_ID_NIL;
+
   return ctx->rc;
 }
 
 grn_tokenizer_query *
-grn_tokenizer_query_open(grn_ctx *ctx, int num_args, grn_obj **args,
+grn_tokenizer_query_open(grn_ctx *ctx,
+                         int num_args,
+                         grn_obj **args,
                          uint32_t normalize_flags)
 {
   grn_obj *flags;
@@ -205,8 +220,8 @@ grn_tokenizer_query_open(grn_ctx *ctx, int num_args, grn_obj **args,
   }
 
   {
-    grn_tokenizer_query * const query =
-        GRN_PLUGIN_MALLOC(ctx, sizeof(grn_tokenizer_query));
+    grn_tokenizer_query *const query =
+      GRN_PLUGIN_MALLOC(ctx, sizeof(grn_tokenizer_query));
     if (!query) {
       GRN_API_RETURN(NULL);
     }
@@ -246,8 +261,8 @@ grn_tokenizer_query_fin(grn_ctx *ctx, grn_tokenizer_query *query)
   if (query->normalized_query) {
     grn_obj_unlink(ctx, query->normalized_query);
   }
-  if (query->query_buf) {
-    GRN_PLUGIN_FREE(ctx, query->query_buf);
+  if (query->data) {
+    GRN_PLUGIN_FREE(ctx, query->data);
   }
 }
 
@@ -304,31 +319,11 @@ grn_tokenizer_query_set_raw_string(grn_ctx *ctx,
                                    const char *string,
                                    size_t string_length)
 {
-  GRN_API_ENTER;
-
-  if (query->query_buf) {
-    GRN_PLUGIN_FREE(ctx, query->query_buf);
-  }
-
-  if (string_length == 0) {
-    query->query_buf = NULL;
-    query->ptr = NULL;
-    query->length = 0;
-    query->need_normalize = GRN_TRUE;
-  } else {
-    query->query_buf = (char *)GRN_PLUGIN_MALLOC(ctx, string_length + 1);
-    if (!query->query_buf) {
-      GRN_PLUGIN_ERROR(ctx, GRN_TOKENIZER_ERROR,
-                       "[tokenizer][query] failed to duplicate query");
-      GRN_API_RETURN(ctx->rc);
-    }
-    grn_memcpy(query->query_buf, string, string_length);
-    query->query_buf[string_length] = '\0';
-    query->ptr = query->query_buf;
-    query->length = string_length;
-  }
-
-  GRN_API_RETURN(ctx->rc);
+  return grn_tokenizer_query_set_data(ctx,
+                                      query,
+                                      string,
+                                      string_length,
+                                      GRN_DB_TEXT);
 }
 
 const char *
@@ -338,9 +333,70 @@ grn_tokenizer_query_get_raw_string(grn_ctx *ctx,
 {
   GRN_API_ENTER;
   if (length) {
-    *length = query->length;
+    *length = query->size;
   }
-  GRN_API_RETURN(query->ptr);
+  GRN_API_RETURN(query->data);
+}
+
+grn_rc
+grn_tokenizer_query_set_data(grn_ctx *ctx,
+                             grn_tokenizer_query *query,
+                             const char *data,
+                             size_t size,
+                             grn_id domain)
+{
+  GRN_API_ENTER;
+
+  if (query->data) {
+    GRN_PLUGIN_FREE(ctx, query->data);
+  }
+
+  if (size == 0) {
+    query->data = NULL;
+  } else {
+    if (grn_type_id_is_text_family(ctx, domain)) {
+      query->data = (char *)GRN_PLUGIN_MALLOC(ctx, size + 1);
+      if (!query->data) {
+        GRN_PLUGIN_ERROR(ctx,
+                         GRN_TOKENIZER_ERROR,
+                         "[tokenizer][query] failed to duplicate query");
+        GRN_API_RETURN(ctx->rc);
+      }
+      grn_memcpy(query->data, data, size);
+      query->data[size] = '\0';
+    } else {
+      query->data = (char *)GRN_PLUGIN_MALLOC(ctx, size);
+      if (!query->data) {
+        GRN_PLUGIN_ERROR(ctx,
+                         GRN_TOKENIZER_ERROR,
+                         "[tokenizer][query] failed to duplicate query");
+        GRN_API_RETURN(ctx->rc);
+      }
+      grn_memcpy(query->data, data, size);
+    }
+  }
+  query->size = size;
+  query->ptr = query->data;
+  query->domain = domain;
+  query->need_normalize = GRN_TRUE;
+
+  GRN_API_RETURN(ctx->rc);
+}
+
+const char *
+grn_tokenizer_query_get_data(grn_ctx *ctx,
+                             grn_tokenizer_query *query,
+                             size_t *size,
+                             grn_id *domain)
+{
+  GRN_API_ENTER;
+  if (size) {
+    *size = query->size;
+  }
+  if (domain) {
+    *domain = query->domain;
+  }
+  GRN_API_RETURN(query->data);
 }
 
 grn_encoding
@@ -462,8 +518,7 @@ grn_tokenizer_query_set_source_column(grn_ctx *ctx,
 }
 
 grn_obj *
-grn_tokenizer_query_get_source_column(grn_ctx *ctx,
-                                      grn_tokenizer_query *query)
+grn_tokenizer_query_get_source_column(grn_ctx *ctx, grn_tokenizer_query *query)
 {
   GRN_API_ENTER;
   GRN_API_RETURN(query->source_column);
@@ -480,16 +535,14 @@ grn_tokenizer_query_set_source_id(grn_ctx *ctx,
 }
 
 grn_id
-grn_tokenizer_query_get_source_id(grn_ctx *ctx,
-                                  grn_tokenizer_query *query)
+grn_tokenizer_query_get_source_id(grn_ctx *ctx, grn_tokenizer_query *query)
 {
   GRN_API_ENTER;
   GRN_API_RETURN(query->source_id);
 }
 
 grn_obj *
-grn_tokenizer_query_get_index_column(grn_ctx *ctx,
-                                     grn_tokenizer_query *query)
+grn_tokenizer_query_get_index_column(grn_ctx *ctx, grn_tokenizer_query *query)
 {
   GRN_API_ENTER;
   GRN_API_RETURN(query->index_column);
@@ -506,8 +559,7 @@ grn_tokenizer_query_set_index_column(grn_ctx *ctx,
 }
 
 grn_obj *
-grn_tokenizer_query_get_options(grn_ctx *ctx,
-                                grn_tokenizer_query *query)
+grn_tokenizer_query_get_options(grn_ctx *ctx, grn_tokenizer_query *query)
 {
   GRN_API_ENTER;
   GRN_API_RETURN(query->options);
@@ -538,8 +590,10 @@ grn_tokenizer_token_fin(grn_ctx *ctx, grn_tokenizer_token *token)
 }
 
 void
-grn_tokenizer_token_push(grn_ctx *ctx, grn_tokenizer_token *token,
-                         const char *str_ptr, unsigned int str_length,
+grn_tokenizer_token_push(grn_ctx *ctx,
+                         grn_tokenizer_token *token,
+                         const char *str_ptr,
+                         unsigned int str_length,
                          grn_token_status status)
 {
   GRN_TEXT_SET_REF(&token->str, str_ptr, str_length);
@@ -555,33 +609,38 @@ grn_tokenizer_tokenized_delimiter_next(grn_ctx *ctx,
                                        unsigned int str_length,
                                        grn_encoding encoding)
 {
-  size_t char_length = 0;
   const char *start = str_ptr;
   const char *current;
   const char *end = str_ptr + str_length;
   const char *next_start = NULL;
-  unsigned int token_length;
-  grn_token_status status;
 
+  int char_length = 0;
   for (current = start; current < end; current += char_length) {
     char_length = grn_charlen_(ctx, current, end, encoding);
     if (char_length == 0) {
       break;
     }
-    if (grn_tokenizer_is_tokenized_delimiter(ctx, current, char_length,
+    if (grn_tokenizer_is_tokenized_delimiter(ctx,
+                                             current,
+                                             (unsigned int)char_length,
                                              encoding)) {
       next_start = str_ptr + (current - start + char_length);
       break;
     }
   }
 
-  token_length = current - start;
+  ptrdiff_t token_length = current - start;
+  grn_token_status status;
   if (current == end) {
     status = GRN_TOKENIZER_LAST;
   } else {
     status = GRN_TOKENIZER_CONTINUE;
   }
-  grn_tokenizer_token_push(ctx, token, start, token_length, status);
+  grn_tokenizer_token_push(ctx,
+                           token,
+                           start,
+                           (unsigned int)token_length,
+                           status);
 
   return next_start;
 }
@@ -593,25 +652,27 @@ grn_tokenizer_next_by_tokenized_delimiter(grn_ctx *ctx,
                                           unsigned int str_length,
                                           grn_encoding encoding)
 {
-  size_t char_length = 0;
   const char *start = str_ptr;
   const char *current;
   const char *end = str_ptr + str_length;
   const char *next_start = NULL;
 
+  int char_length = 0;
   for (current = start; current < end; current += char_length) {
     char_length = grn_charlen_(ctx, current, end, encoding);
     if (char_length == 0) {
       break;
     }
-    if (grn_tokenizer_is_tokenized_delimiter(ctx, current, char_length,
+    if (grn_tokenizer_is_tokenized_delimiter(ctx,
+                                             current,
+                                             (unsigned int)char_length,
                                              encoding)) {
       next_start = str_ptr + (current - start + char_length);
       break;
     }
   }
 
-  grn_token_set_data(ctx, token, start, current - start);
+  grn_token_set_data(ctx, token, start, (int)(current - start));
   if (current == end) {
     grn_token_set_status(ctx, token, GRN_TOKEN_LAST);
   } else {
@@ -622,9 +683,11 @@ grn_tokenizer_next_by_tokenized_delimiter(grn_ctx *ctx,
 }
 
 grn_rc
-grn_tokenizer_register(grn_ctx *ctx, const char *plugin_name_ptr,
+grn_tokenizer_register(grn_ctx *ctx,
+                       const char *plugin_name_ptr,
                        unsigned int plugin_name_length,
-                       grn_proc_func *init, grn_proc_func *next,
+                       grn_proc_func *init,
+                       grn_proc_func *next,
                        grn_proc_func *fin)
 {
   grn_expr_var vars[3];
@@ -638,10 +701,15 @@ grn_tokenizer_register(grn_ctx *ctx, const char *plugin_name_ptr,
       grn_proc_create() registers a plugin to the database which is associated
       with `ctx'. A returned object must not be finalized here.
      */
-    grn_obj * const obj = grn_proc_create(ctx, plugin_name_ptr,
-                                          plugin_name_length,
-                                          GRN_PROC_TOKENIZER,
-                                          init, next, fin, 3, vars);
+    grn_obj *const obj = grn_proc_create(ctx,
+                                         plugin_name_ptr,
+                                         (int)plugin_name_length,
+                                         GRN_PROC_TOKENIZER,
+                                         init,
+                                         next,
+                                         fin,
+                                         3,
+                                         vars);
     if (obj == NULL) {
       GRN_PLUGIN_ERROR(ctx, GRN_TOKENIZER_ERROR, "grn_proc_create() failed");
       return ctx->rc;
@@ -651,9 +719,7 @@ grn_tokenizer_register(grn_ctx *ctx, const char *plugin_name_ptr,
 }
 
 grn_obj *
-grn_tokenizer_create(grn_ctx *ctx,
-                     const char *name,
-                     int name_length)
+grn_tokenizer_create(grn_ctx *ctx, const char *name, int name_length)
 {
   grn_obj *tokenizer;
 
@@ -669,12 +735,13 @@ grn_tokenizer_create(grn_ctx *ctx,
                               NULL);
   if (!tokenizer) {
     if (name_length < 0) {
-      name_length = strlen(name);
+      name_length = (int)strlen(name);
     }
     GRN_PLUGIN_ERROR(ctx,
                      GRN_TOKENIZER_ERROR,
                      "[tokenizer][create] failed to create: <%.*s>",
-                     name_length, name);
+                     name_length,
+                     name);
   }
 
   GRN_API_RETURN(tokenizer);
